#from ..config import iamPolicyAttributes, arnServiceMap, awsAccount
import iam_check.config as config
import enum
import logging
import json
import re


LOGGER = logging.getLogger('iam-policy-validator-for-terraform')

class TerraformState:
    def __init__(self, **kwargs) -> None:
        """Not implemented since it only supports planned changes"""
        self.format_version = None
        self.terraform_version = None
        self.values = None 

        for arg, value in kwargs.items():
            if arg == 'format_version':
                self.format_version = value
            elif arg == 'terraform_version':
                self.terraform_version = value
            else:
                setattr(self, arg, value)
    def __str__(self):
        return json.dumps(vars(self), indent=4)

    def getValue (self, key):
        raise NotImplementedError('Not implemented as we only support planned changes')

class TerraformPlan:
    def __init__(self, **kwargs) -> None:
        self.format_version = None
        self.prior_state = None
        self.configuration = None
        self.planned_values = None
        self.proposed_unknown = None
        self.variables = {}
        self.resource_changes = []
        self.output_changes = []
        self.ref_map = {}

        for arg, value in kwargs.items():
            if arg == 'format_version':
                self.format_version = value
            elif arg == 'prior_state':
                self.prior_state = TerraformState(**value)
            elif arg == 'configuration':
                self.configuration = TerraformConfig(**value)
            elif arg == 'planned_values':
                if isinstance(value, TerraformValues):
                    self.planned_values = value
                else:
                    self.planned_values = TerraformValues(**value)
            elif arg == 'proposed_unknown':
                self.proposed_unknown = TerraformValues(**value)
            elif arg == 'variables':
                self.variables = value
            elif arg == 'resource_changes':
                self.resource_changes = [TerraformResourceChange(**r) for r in value]
            elif arg == 'output_changes':
                self.output_changes = value
    
    def __str__(self) -> str:
        obj = {}
        obj['format_version'] = self.format_version
        obj['variables'] = self.variables
        obj['resource_changes'] = [json.loads(str(r)) for r in self.resource_changes]
        obj['output_changes'] = self.output_changes
        
        if self.prior_state is not None:
            obj['prior_state'] = json.loads(str(self.prior_state))
        if self.configuration is not None:
            obj['configuration'] = json.loads(str(self.configuration))
        if self.planned_values is not None:
            obj['planned_values'] = json.loads(str(self.planned_values))
        if self.proposed_unknown is not None:
            obj['proposed_unknown'] = json.loads(str(self.proposed_unknown))
        return json.dumps(obj, indent=4)

    def findPolicies(self):
        logging.debug('generating a list of policies in plan')
        policies = {}
        resources = self.listResources()
        for r in resources:
            resourceType = r.split('.')[-2]
            if resourceType not in config.iamPolicyAttributes:
                continue
            
            attributes = config.iamPolicyAttributes[resourceType]
            if isinstance(attributes, str):
                attributes = [attributes]
            for attribute in attributes:
                # check if attribute is a base one or inside a block
                if '.' not in attribute:
                    ref = f'{r}.{attribute}'
                    try:
                        policy = self.getValue(ref)
                        if policy is None or policy == "":
                            LOGGER.info(f'No policy found at: {ref}')
                        else:
                            policies[ref] = policy
                    except KeyError as e:
                        LOGGER.info(f'No policy found at: {ref}')
                        continue
                else:
                    block, key = attribute.split('.')
                    block_ref = f'{r}.{block}'
                    try:
                        data = self.getValue(block_ref)
                    except KeyError as e:
                        LOGGER.info(f'No policy found at: {ref}')
                        continue
                    
                    if not isinstance (data, list):
                        ref = f'{block_ref}.key'
                        policy = data[key]
                        if policy is None or policy == "":
                            LOGGER.info(f'No policy found at: {ref}')
                        else:
                            policies[ref] = policy
                    else:
                        for index, block in enumerate(data):
                            ref = f'{block_ref}.{index}.{key}'
                            self.ref_map[f'{block_ref}.{index}'] = r
                            if "policy" in block: 
                                policy = block[key]
                                if policy is None or policy == "":
                                    LOGGER.info(f'No policy found at: {ref}')
                                else:
                                    policies[ref] = policy
                            else:
                                LOGGER.info(f'No policy found at: {ref}')
        for ref in policies.keys():
            LOGGER.debug(f'found policy at {ref}')
        
        return policies
    def getResourceName(self, ref):
        if isinstance(ref, TerraformResource):
            resource =  ref
            ref = resource.address
        else:
            mapped_ref = self.ref_map.get(ref)
            if mapped_ref is not None:
                resource = self.getValue(mapped_ref)
            else:
                resource = self.getValue(ref)

        if resource.type not in config.arnServiceMap:
            raise TypeError(f'Add resource type {resource.type} in the configuration arnServiceMap')
        key = config.arnServiceMap[resource.type]
        terraformKey = key
        default = None
        if '?' in key:
            terraformKey, default = key.split('?')
        try:
            name = self.getValue(f'{ref}.{terraformKey}')
        except KeyError as e:
            if default is None:
                raise e
            name = default
        return name

    def getFakeArn(self, ref):
        if isinstance(ref, TerraformResource):
            resource =  ref
            ref = resource.address
        else:
            resource = self.getValue(ref)

        if resource.type not in config.arnServiceMap:
            raise TypeError(f'Add resource type {resource.type} in the configuration arnServiceMap')
        arn_pattern = config.arnServiceMap[resource.type]
        arnComponents = arn_pattern.split(':')
        arnComponents[4] = config.awsAccount
        arn_pattern = ':'.join(arnComponents)
        arn_keys = re.findall('{([^}]+)}', arn_pattern)
        print(arn_keys)
        arn_dict = {}
        for key in arn_keys:
            terraformKey = key
            default = None
            if '?' in key:
                terraformKey, default = key.split('?')
            try:
                arn_dict[key] = self.getValue(f'{ref}.{terraformKey}')
                print(terraformKey)
            except KeyError as e:
                if default is None:
                    raise e
                arn_dict[key] = default
            print(arn_dict[key])
        return arn_pattern.format(**arn_dict)

    def getValue(self, ref):
        """ interpolate a given resource address and attribute based on computed values and return the results 
        """
        found = False
        result = None
        # first get the original values
        try: 
            result = self.prior_state.getValue(ref)
            found = True
        except: pass

        # then get the planned value
        try: 
            result = self.planned_values.getValue(ref)
            found = True
        except: pass

        if found is False:
            raise KeyError(f'Invalid Key: {ref}')

        if isinstance(result, str):
            return self.parserHclString(result)
        return result

    def listResources(self):
        "created or updated resources in plan"
        return self.planned_values.listResources()
    
    def parserHclString(self, input):
        "This function will only attempt direct lookups"
        interpolation = '${}'
        directive = '${}'
        result = ""

        class strFsm(enum.Enum):
            parseString = enum.auto()
            startInterpolation = enum.auto()
            startDirective = enum.auto()
            parseInterpolation = enum.auto()
            parseDirective = enum.auto()
            
        state = strFsm.parseString

        for char in input:
            if state is strFsm.parseString:
                if char == interpolation[0]:
                    state = strFsm.startInterpolation
                elif char == directive[0]:
                    state = strFsm.startDirective
                else:
                    result +=char
            elif state is strFsm.startInterpolation: 
                if char == interpolation[0]:
                    state = strFsm.parseString
                    result +=char
                elif char == interpolation[1]:
                    state = strFsm.parseInterpolation
                    expression = ""
                else:
                    state = strFsm.parseString
                    result += interpolation[0] + char

            elif state is strFsm.startDirective:
                if char == directive[0]:
                    state = strFsm.parseString
                    result +=char
                elif char == directive[1]:
                    state = strFsm.parseDirective
                    expression = ""
                else:
                    state = strFsm.parseString
                    result += directive[0] + char

            elif state is strFsm.parseInterpolation:
                if char == interpolation[2]:
                    state = strFsm.parseString
                    try:
                        result += self.getValue(expression)
                    except:
                        LOGGER.warning(f'string interpolation not available: {expression}')
                        expression = '${' + expression + '}'
                        result += expression
                else:
                    expression += char
            elif state is strFsm.parseDirective:
                if char == directive[2]:
                    state = strFsm.parseString         
                    expression = '${' + expression + '}'
                    result += expression
                    LOGGER.warning(f'Cannot process string directive: {directive}')
                else:
                    expression += char
            else:
                raise RuntimeError(f'FSM for parsing HCL string reached an unknown state: {state}')
        return result


class TerraformValues:
    def __init__(self, **kwargs) -> None:
        self.outputs = None
        self.root_module = None

        for arg, value in kwargs.items():
            if arg == 'outputs':
                self.outputs = value
            elif arg == 'root_module':
                if isinstance(value, TerraformModule):
                    self.root_module = value
                else:
                    self.root_module = TerraformModule(**value)
            else:
                raise ValueError(f'TerraformVaules: Unknown parameter: {arg}')
    def __eq__(self, o: object) -> bool:
        return vars(self) == vars(o)
    
    def __str__(self):
        obj = {k: json.loads(str(v)) for k,v in vars(self).items() if v is not None}
        return json.dumps(obj, indent=4)
    
    def getValue(self, key):
        return self.root_module.getValue(key)
    
    def addResource(self, r):
        self.root_module.addResource(r)

    def addChildModule(self, m):
        self.root_module.addChildModule(m)

    def listResources(self):
        return self.root_module.listResources()

class TerraformModule:
    def __init__(self, **kwargs) -> None:
        self.address = None
        self.resources = []
        self.child_modules = []

        for arg, value in kwargs.items():
            if arg == "address":
                self.address = value
            elif arg == "resources":
                for resource in value:
                    if isinstance(resource, TerraformResource):
                        self.resources.append(resource)
                    else:
                        self.resources.append( TerraformResource(**resource) )
            elif arg == "child_modules":
                for child in value:
                    self.child_modules.append(TerraformModule(**child))
            else:
                raise ValueError(f'Unknown parameter: {arg}')
            
    def __eq__(self, o: object) -> bool:
        if self.address != o.address:
            return False
        if self.resources != o.resources:
            return False
        if self.child_modules != o.child_modules:
            return False
        return True
    
    def __str__(self):
        obj = {}

        if self.address is not None:
            obj['address'] = self.address

        obj['resources'] = []
        for r in self.resources:
            data = json.loads(str(r))
            obj['resources'].append(data)
        
        obj['child_modules'] = []
        for m in self.child_modules:
            data = json.loads(str(m))
            obj['child_modules'].append(data)
        return json.dumps(obj, indent=4)

    def addResource(self, r):
        if not isinstance(r, TerraformResource):
            raise ValueError('Must be a TerraformResource')
        self.resources.append(r)
    
    def addChildModule(self, m):
        if not isinstance(m, TerraformModule):
            raise ValueError('Must be a TerraformModule')
        self.child_modules.append(m)

    def getValue(self, key):
        if key == self.address:
            return self
        
        for r in self.resources:
            if r.address == key:
                return r
            if key.startswith(f'{r.address}.'):
                return r.getValue(key)
            
        for m in self.child_modules:
            if m.address == key:
                return key
            if key.startswith(f'{m.address}.'):
                return m.getValue(key)
        raise KeyError(f'Invalid terraform address: {key}')

    def listResources(self):
        result = [r.address for r in self.resources]
        for m in self.child_modules:
            result = result + m.listResources()
        return result

class TerraformResource:
    def __init__(self, **kwargs) -> None:
        self.address = None
        self.mode = None        
        self.type = None
        self.name = None
        self.index = None
        self.change = None
        self.provider_name = None
        self.schema_version = None
        self.values = {}
        self.sensitive_values = {}

        for arg, value in kwargs.items():
            if arg == 'address':
                self.address = value
            elif arg == 'mode':
                if value not in ['managed', 'data']:
                    raise ValueError(f'TerraformResource: Invalid mode: {value}')
                self.mode = value
            elif arg == 'type':
                self.type = value
            elif arg == 'name':
                self.name = value
            elif arg == 'index':
                self.index = value
            elif arg == 'change':
                self.change = TerraformChange(**value)
            elif arg == 'provider_name':
                self.provider_name = value
            elif arg == 'schema_version':
                self.schema_version = value
            elif arg == 'values':
                self.values = value
            elif arg == 'sensitive_values':
                self.sensitive_values = value
            else:
                raise ValueError(f'TerraformResource: Invalid parameter: {arg}')

    def __eq__(self, o: object) -> bool:
        return vars(self) == vars(o)

    def __str__(self):
        obj = {k:v for k,v in vars(self).items() if v is not None}
        if self.change is not None:
            obj['change'] = json.loads(str(self.change))
        return json.dumps(obj, indent=4)

    def getValue(self, key):
        if key.startswith(self.address):
            key = key[len(self.address)+1:]

        if key in self.values:
            return self.values[key]

        if key in self.sensitive_values:
            return self.sensitive_values[key]
        raise KeyError(f'TerraformResources: {self.address}.{key}')
    
    def setValue(self, key, value, sensitive = False):
        if sensitive:
            self.sensitive_values[key] = value
        else:
            self.values[key] = value

class TerraformChange:
    def __init__(self, **kwargs) -> None:
        """Not implemented since it is not needed to find policy strings"""
        pass

    def __str__(self):
        obj = vars(self)
        return json.dumps(obj, indent=4)

class TerraformResourceChange:
    def __init__(self, **kwargs) -> None:
        """Not implemented since it is not needed to find policy strings"""
        pass

    def __str__(self):
        obj = {k: json.loads(str(v)) for k,v in vars(self).items() if v is not None}
        return json.dumps(obj, indent=4)
        
class TerraformConfig:
    def __init__(self, **kwargs) -> None:
        """Not implemented since it is not needed to find policy strings"""
        self.provider_configs = None
        self.root_module = None

    def __str__(self):
        obj = {k: json.loads(str(v)) for k,v in vars(self).items() if v is not None}
        return json.dumps(obj, indent=4)

def plan_from_stdout(input):
    """attempt to convert output from `terraform show` into a data structure. Only compatiable with 0.11 and earlier versions"""
    data = {}
    resource = None
    operation = None
    operators = ['+', '-', '~', '+/-']

    for line in input.split('\n'):
        # remove color encoding
        line = re.sub(r'\033\[[0-9;]*m', '', line)
        line = line.strip()

        #blank lines indicate done processing previous resource
        if len(line) == 0:
            resource = None
            operation = None
            continue

        # determine the operation on the given resource
        key, value = line.split(' ', 1)
        if operation is None:
            if key not in operators:
                raise ValueError('Unkown operations: {line}')
            operation = key
            resource = value
            if operation != '-':
                data[resource] = {}
            continue
        elif operation == '-':
            continue

        
        key = key.strip()
        key = key.rstrip(':')
        value = value.strip()

        if operation in ['+/-', '~']:
            value = value.split(' => ')[1]
        
        # if it is a string it needs to be decoded
        if value[0] == '"':
            value = value.strip('"')
            value = value.encode('utf-8').decode('unicode_escape')

        data[resource][key] = value

    root_module = TerraformModule()

    for resource, attributes in data.items():
        r = TerraformResource()
        r.address = resource
        r.type = resource.split('.')[-2]
        r.name = resource.split('.')[-1]

        for k, v in attributes.items():
            if v == "<computed>":
                continue
            r.values[k] = v

        root_module.addResource(r)

    values = TerraformValues(root_module = root_module)
    plan = TerraformPlan(planned_values = values)

    return plan



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import boto3
from botocore.config import Config

config = Config(
	retries={
		# this number was chosen arbitrarily, tweak as necessary
		'max_attempts': 30,
		'mode': 'standard'
	}
)


def get_account_and_partition(region):
	"""
	Pull the account and partition from the credentials used to execute the validator
	"""

	sts_client = build('sts', region)
	identity = sts_client.get_caller_identity()
	account_id = identity['Account']

	parts = identity['Arn'].split(':')
	partition = parts[1]

	return account_id, partition


def build(service_name, region_name, client_config=None):
	if client_config is None:
		client_config = config
	session = boto3.Session(profile_name=profile_name, region_name=region_name)
	return session.client(service_name, config=client_config)


profile_name = None


def set_profile(profile):
	global profile_name
	profile_name = profile


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from . import default_to_json
from iam_check.tools import regex_patterns

default_finding_types_that_are_blocking = ['ERROR', 'SECURITY_WARNING']


class Reporter:
	"""
	Determines what findings should be reported to the end user based on parameters provided when starting validation.
	"""

	def __init__(self, findings_to_ignore, finding_types_that_are_blocking, allowed_external_principals):
		self.blocking_findings = []
		self.nonblocking_findings = []
		self.findings_to_ignore = findings_to_ignore
		self.finding_types_that_are_blocking = finding_types_that_are_blocking
		self.allowed_external_principals = allowed_external_principals

	def build_report_from(self, findings):
		self._filter_overridden_findings(findings)
		return Report(self.blocking_findings, self.nonblocking_findings)

	def _filter_overridden_findings(self, findings):
		for finding in findings.errors + findings.security_warnings + findings.warnings + findings.suggestions:
			overridden = self._is_finding_ignored(finding)
			if overridden:
				continue

			overridden = self._is_external_principal_allowed(finding)
			if overridden:
				continue

			self._classify_as_blocking_or_non_blocking(finding)

	def _is_finding_ignored(self, finding):
		if self.findings_to_ignore is None:
			return False

		is_ignored = any([finding_to_ignore.matches(finding) for finding_to_ignore in self.findings_to_ignore])
		if is_ignored:
			return True

		return False

	def _is_external_principal_allowed(self, finding):
		if self.allowed_external_principals is None:
			return False

		is_allowed = any([principal_to_allow.matches(finding) for principal_to_allow in self.allowed_external_principals])
		if is_allowed:
			return True

		return False

	def _classify_as_blocking_or_non_blocking(self, finding):
		# if none is present it overrides all others
		if 'NONE' in self.finding_types_that_are_blocking:
			self.nonblocking_findings.append(finding)
			return

		if finding.findingType.upper() in self.finding_types_that_are_blocking:
			self.blocking_findings.append(finding)
		else:
			self.nonblocking_findings.append(finding)


class ResourceOrCodeFindingToIgnore:
	def __init__(self, value):
		self.value = value

	def matches(self, finding):
		return finding.resourceName.lower() == self.value.lower() or \
				finding.code.lower() == self.value.lower()

	def __eq__(self, other):
		if not isinstance(other, ResourceOrCodeFindingToIgnore):
			return False

		return self.value == other.value


class ResourceAndCodeFindingToIgnore:
	def __init__(self, resource_name, code):
		self.resource_name = resource_name
		self.code = code

	def matches(self, finding):
		return finding.resourceName.lower() == self.resource_name.lower() and \
				finding.code.lower() == self.code.lower()

	def __eq__(self, other):
		if not isinstance(other, ResourceAndCodeFindingToIgnore):
			return False

		return self.resource_name == other.resource_name and \
				self.code == other.code


def _get_principal_from_finding(finding):
	if finding.code != 'EXTERNAL_PRINCIPAL':
		return None

	principal = finding.details.get('principal', {})
	aws_principal = principal.get('AWS')
	if aws_principal is not None:
		return aws_principal

	federated_principal = principal.get('Federated')
	if federated_principal is not None:
		return federated_principal

	return principal.get('CanonicalUser')


class AllowedExternalPrincipal:
	def __init__(self, principal):
		self.principal = principal

	def matches(self, finding):
		principal = _get_principal_from_finding(finding)
		if principal is None:
			return False

		match = regex_patterns.generic_arn_pattern.match(principal)
		if match is None:
			# the principal may or may not be an ARN, if it's not an ARN, compare the raw values
			return principal == self.principal

		# if principal is an ARN, grab the account ID from the ARN to compare
		account_id = match.group(1)
		return account_id == self.principal

	def __eq__(self, other):
		if not isinstance(other, AllowedExternalPrincipal):
			return False

		return self.principal == other.principal


class AllowedExternalArn:
	def __init__(self, arn):
		self.arn = arn

	def matches(self, finding):
		principal = _get_principal_from_finding(finding)
		if principal is None:
			return False

		return principal == self.arn

	def __eq__(self, other):
		if not isinstance(other, AllowedExternalArn):
			return False

		return self.arn == other.arn


class Report:
	def __init__(self, blocking_findings, nonblocking_findings):
		self.blocking_findings = blocking_findings
		self.nonblocking_findings = nonblocking_findings

	def has_blocking_findings(self):
		return len(self.blocking_findings) > 0

	def to_json(self):
		return {
			'BlockingFindings': [vars(finding) for finding in self.blocking_findings],
			'NonBlockingFindings': [vars(finding) for finding in self.nonblocking_findings]
		}

	def print(self):
		report = self.to_json()
		report_as_json_string = self._to_json_string(report)
		print(report_as_json_string)

	@staticmethod
	def _to_json_string(obj):
		return json.dumps(obj, default=default_to_json, indent=4)


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
from argparse import ArgumentTypeError
from botocore.config import Config
from botocore.exceptions import InvalidRegionError
from botocore.utils import validate_region_name

from .client import build
# from cfn_policy_validator.application_error import ApplicationError

def validate_region(region):
	try:
		# this call validates that the region name is valid, but does not validate that the region actually exists
		validate_region_name(region)
	except InvalidRegionError:
		raise ArgumentTypeError(f'Invalid region name: {region}.')
	return region

def validate_credentials(region):
	# run a test to validate the provided credentials
	# create our own config here to control retries and fail fast if credentials are invalid
	sts_client = build('sts', region, client_config=Config(retries={'mode': 'standard', 'max_attempts': 2}))
	sts_client.get_caller_identity

def validate_finding_types_from_cli(value):
	"""
	Validate that the finding types provided are valid finding types.
	"""

	finding_types = value.split(',')
	finding_types = validate_finding_types(finding_types)

	return finding_types


def validate_finding_types(finding_types):
	if finding_types is None:
		return finding_types

	finding_types = [finding_type.strip() for finding_type in finding_types]
	finding_types = [finding_type.upper() for finding_type in finding_types]

	for finding_type in finding_types:
		if finding_type not in ['ERROR', 'SECURITY_WARNING', 'SUGGESTION', 'WARNING', 'NONE']:
			raise ArgumentTypeError(f"Invalid finding type: {finding_type}.")

	return finding_types


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
class AccountConfig:
	def __init__(self, partition, region, account_id):
		self.Account = account_id
		self.Region = region
		self.Partition = partition


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import json

from . import default_to_json


class Finding:
	def __init__(self, message, finding_type, policy_name, resource_name, details, code):
		self.findingType = finding_type
		self.code = code
		self.message = message
		self.resourceName = resource_name
		self.policyName = policy_name
		self.details = details


class Findings:
	"""
	Build a findings object from the Access Analyzer response, wrapping the raw findings from Access Analyzer so that
	consumers can parse the output in a standardized manner.
	"""

	def __init__(self):
		self.errors = []
		self.security_warnings = []
		self.warnings = []
		self.suggestions = []

	def add_trust_policy_finding(self, findings, resource_name):
		for raw_finding in findings:
			message = 'Trust policy allows access from external principals.'
			finding = Finding(
				message=message,
				finding_type='SECURITY_WARNING',
				policy_name='TrustPolicy',
				resource_name=resource_name,
				details=raw_finding,
				code='EXTERNAL_PRINCIPAL'
			)

			self.security_warnings.append(finding)

	def add_validation_finding(self, findings, resource_name, policy_name):
		for raw_finding in findings:
			finding_type = raw_finding['findingType']

			finding = Finding(
				message=raw_finding['findingDetails'],
				finding_type=finding_type,
				policy_name=policy_name,
				resource_name=resource_name,
				details=raw_finding,
				code=raw_finding['issueCode']
			)

			if finding_type == 'ERROR':
				self.errors.append(finding)
			elif finding_type == 'SECURITY_WARNING':
				self.security_warnings.append(finding)
			elif finding_type == 'SUGGESTION':
				self.suggestions.append(finding)
			elif finding_type == 'WARNING':
				self.warnings.append(finding)

	def add_policy_analysis_finding(self, findings, resource_name, policy_name):
		for raw_finding in findings:
			finding_type = raw_finding.get('findingType')
			finding = Finding(
				message=raw_finding.get('message'),
				finding_type=finding_type,
				policy_name=policy_name,
				resource_name=resource_name,
				details=raw_finding.get('response'),
				code=raw_finding.get('code')
			)
			if finding_type == 'ERROR':
				self.errors.append(finding)
			elif finding_type == 'SECURITY_WARNING':
				self.security_warnings.append(finding)

	def add_external_principal_finding(self, findings, resource_name, policy_name):
		for raw_finding in findings:
			finding = Finding(
				message='Resource policy allows access from external principals.',
				finding_type='SECURITY_WARNING',
				policy_name=policy_name,
				resource_name=resource_name,
				details=raw_finding,
				code='EXTERNAL_PRINCIPAL'
			)

			self.security_warnings.append(finding)

	def to_json(self):
		return json.dumps(self, default=default_to_json, indent=4)


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import re

# matches an ARN generically and roughly answers "is this an ARN or not?"
# captures the account ID
generic_arn_pattern = re.compile(r"^arn:aws[a-z0-9-]*:.*:.*:(.*):.*$")


# this looks for values within a string that are surrounded by dollar signs and brackets
# the values must not start with ! (!${}) - as this represents a raw value
# used by Fn::Sub
# e.g. ${MyValue} -> MyValue
fn_sub_variables = re.compile(r'\$\{([^!].*?)\}')


# looks for dynamic ssm regex of the form {{resolve:ssm:reference-key:version}} or {{resolve:ssm:reference-key}}
# captures reference-key and version (if it exists)
dynamic_ssm_reference_regex = re.compile(r'({{resolve:ssm:([a-zA-Z0-9_\.\-\/]+):?(\d+)?}})')


import json
import logging
from iam_check import application_error
from iam_check.client import build
from .findings import Findings
from . import iamPolicy
import iam_check.config as config
from botocore.exceptions import ClientError
from botocore.config import Config

LOGGER = logging.getLogger('iam-policy-validator-for-terraform')
ACTIONS_MAX_ITEMS = 100
POLICY_ANALYSIS_PREFIX = 'policy-analysis-'

class Validator:
    def __init__(self, account_id, region, partition):
        self.findings = Findings()
        self.access_analyzer_name = 'AnalyzerCreatedByTfIAMPolicyValidator'
        self.analyzer_arn = None
        self.client = build('accessanalyzer', region)
		# preview builders are used to build the access preview configuration for an individual resource type
		# a preview builder must be added to add support for access previews for a given resource
#         self.preview_builders = {
# 			'AWS::SQS::Queue': SqsQueuePreviewBuilder(account_id, region, partition),
# 			'AWS::KMS::Key': KmsKeyPreviewBuilder(account_id, region, partition),
# 			'AWS::S3::AccessPoint': S3SingleRegionAccessPointPreviewBuilder(account_id, region, partition),
# 			'AWS::S3::MultiRegionAccessPoint': S3MultiRegionAccessPointPreviewBuilder(account_id, partition),
# 			'AWS::S3::Bucket': S3BucketPreviewBuilder(region, partition),
# 			'AWS::IAM::Role::TrustPolicy': RoleTrustPolicyPreviewBuilder(account_id, partition),
# 			'AWS::SecretsManager::Secret': SecretsManagerSecretPreviewBuilder(account_id, region, partition)
# 		}
		# maps the resource type to the parameter for validate_policy that enables service specific policy validation
		# not all services have service specific policy validation.  The names may be identical for now, but we don't
		# want to rely on that
		#to move this to config file
        self.maximum_number_of_access_preview_attempts = 150
        
    def run(self, plan):
        policies = plan.findPolicies()
        for ref, policy in policies.items():
            LOGGER.info(f'check policy at: {ref}')
            policy_resource_type=ref.split('.')[0]
            policy_name = '.'.join(ref.split('.')[0:-1])
            resource_name = plan.getResourceName(policy_name)
            p = iamPolicy.Policy(json=policy)
            LOGGER.info(f'start checking policy:{p}')
            policyType='IDENTITY_POLICY'
            for statement in p.getStatement():
                if statement.getPrincipal() != None or statement.getNotPrincipal() != None:
                    policyType='RESOURCE_POLICY'
                    continue
            if policy_resource_type not in config.validatePolicyResourceType :
                response = self.client.validate_policy(policyDocument=str(p),policyType=policyType)
            else:
                policy_resource_type = config.validatePolicyResourceType[policy_resource_type]
                response = self.client.validate_policy(
                    policyDocument=str(p),
                    policyType=policyType,
                    validatePolicyResourceType=policy_resource_type
                )
            validation_findings = response['findings']
            self.findings.add_validation_finding(validation_findings, resource_name, policy_name)

def get_policy_type(resource_type, resource_attribute_name = None):
        # https://registry.terraform.io/providers/hashicorp/aws/latest/docs
        if resource_type in [
            "aws_iam_group_policy",
            "aws_iam_policy",
            "aws_iam_role",
            "aws_iam_role_policy",
            "aws_iam_user_policy"
        ]:
            if resource_type == "aws_iam_role" and resource_attribute_name == "assume_role_policy":
                return "RESOURCE_POLICY"
            return "IDENTITY_POLICY"
        else:
            return "RESOURCE_POLICY"

class PolicyAnalysis:
    def __init__(self, region):
        self.findings = Findings()
        self.client = build('accessanalyzer', region, client_config=Config(
            retries={
                # this number was chosen arbitrarily, tweak as necessary
                'max_attempts': 3,
                'mode': 'standard'
            }
        ))
        self.findings = Findings()
        self.identity_policy_cache = {}
        self.resource_policy_cache = {}
    
    
    def _handle_response(self, response, resource_name, policy_name, operation_name):
        """
        Builds a list of raw findings based on the API response
        """
        # check-access passes in a list as it does batching and calling for more than 50 actions
        if isinstance(response, list):
            findings = [self._build_policy_analysis_finding(r, operation_name) for r in response if r.get('result') != 'PASS']
        elif response.get('result') != 'PASS':
            findings = [self._build_policy_analysis_finding(response, operation_name)]
        else:
            findings = []
        self.findings.add_policy_analysis_finding(findings, resource_name, policy_name)

    def _build_policy_analysis_finding(self, response, operation_name):
        """
        Create a raw finding for non 'PASS' results
        """
        code = f'{POLICY_ANALYSIS_PREFIX}{operation_name}'
        response_code = response['ResponseMetadata']['HTTPStatusCode']
        if response_code != 200:
            # error response shape https://boto3.amazonaws.com/v1/documentation/api/latest/guide/error-handling.html
            # response['Error']['Code'] returns literal exception name such as ValidationException, UnprocessableEntityException
            # Raise error for non 400 errors
            if (response_code < 400 or response_code > 499):
                raise application_error(response.get('message'))
            # Add finding for 400 errors
            else:
                code += response['Error']['Code']
                finding_type = "ERROR"    
        else:
            finding_type = "SECURITY_WARNING"
        response_no_metadata = response.copy()
        rawFinding = {
            'message': response.get('message'), 
            'findingType': finding_type,
            'response': response_no_metadata,
            'code': code
        }
        del response_no_metadata['ResponseMetadata']
        return rawFinding

    def _call_api(self, policy_as_json, policy_type):
        pass

    def _get_policy_type():
        pass
    
    def run(self, plan):
        policies = plan.findPolicies()
        for ref, policy_str in policies.items():
            LOGGER.info(f'check policy at: {ref}')
            policy_resource_type=ref.split('.')[0]
            policy_attribute_type=ref.split('.')[-1]
            policy_name = '.'.join(ref.split('.')[0:-1])
            resource_name = plan.getResourceName(policy_name)
            LOGGER.info(f'start checking policy:{policy_str}')
            policy_type = get_policy_type(policy_resource_type, policy_attribute_type) 
            cache = self.identity_policy_cache if policy_type == "IDENTITY_POLICY" else self.resource_policy_cache
            response = cache.get(policy_str)
            if response is None:
                response = self._call_api(policy_str, policy_type)
                cache[policy_str] = response
            self._handle_response(response, resource_name, policy_name, self.operation_name)


class Comparator(PolicyAnalysis):

    def __init__(self, region, reference_policy, reference_policy_type):
        PolicyAnalysis.__init__(self, region)
        self.reference_policy = reference_policy
        self.operation_name = "CheckNoNewAccess"
        self.reference_policy_type = reference_policy_type
    
    def _call_api(self, policy, policy_type):
        if self.reference_policy_type == "identity" and policy_type != "IDENTITY_POLICY":
            return []
        if self.reference_policy_type == "resource" and policy_type != "RESOURCE_POLICY":
            return []
        try:
            response = self.client.check_no_new_access(
                policyType=policy_type,
                existingPolicyDocument=self.reference_policy,
                newPolicyDocument=policy,
            )
        except ClientError as error:
            return error.response

        return response

class AccessChecker(PolicyAnalysis):

    def __init__(self, region, actions):
        PolicyAnalysis.__init__(self, region)
        self.accesses = []
        for i in range(0, len(actions), ACTIONS_MAX_ITEMS): 
            access = [{
                "actions": actions[i:min(ACTIONS_MAX_ITEMS + i, len(actions))]
            }]
            self.accesses.append(access)
        self.operation_name = "CheckAccessNotGranted"

    def _call_api(self, policy, policy_type):
        responses = []
        failed_response = {
            "result": "FAIL",
            "message": "The policy document grants access to perform one or more of the listed actions.",
            "reasons": []
        }
        for access in self.accesses:
            LOGGER.info(f'Batching actions {access}')
            try:
                response = self.client.check_access_not_granted(
                    policyType=policy_type,
                    policyDocument=policy,
                    access=access
                )
            except ClientError as error:
                response = error.response
            if response.get('result') == 'FAIL':
                reasons = response.get('reasons')
                if reasons is not None:
                    for r in reasons:
                        r['accessInput'] = access
                    failed_response['reasons'].extend(reasons)
                else:
                    failed_response['reasons'].append({
                        'accessInput': access
                    })
                failed_response['ResponseMetadata'] = response['ResponseMetadata']
            else:
                response['accessInput'] = access
                responses.append(response)
        if failed_response.get('ResponseMetadata'): # There were fail responses
            responses.append(failed_response)
        return responses


#!/usr/bin/env python3

import argparse
import logging
import json
import sys
import traceback

import iam_check.config as config
from .client import get_account_and_partition, set_profile
from .parameters import validate_region, validate_finding_types_from_cli, validate_credentials
from .argument_actions import ParseFindingsToIgnoreFromCLI, ParseAllowExternalPrincipalsFromCLI, ParseListFromCLI
from .lib import _load_json_file, iamcheck_AccessAnalyzer, account_config, tfPlan
from .lib.reporter import default_finding_types_that_are_blocking, Reporter

# Global Variables
LOGGER = logging.getLogger('iam-policy-validator-for-terraform')

def main():
    global policy_checks
    opts = cli_parse_opts()
    validate_from_cli(opts)

def validate_from_cli(opts):
    account_id, partition = get_account_and_partition(opts.region)
    account = account_config.AccountConfig(partition, opts.region, account_id)
    findings = []
    LOGGER.debug(f'Validating terraform plan file: {opts.template_path}')
    with open(opts.template_path, 'r') as fh:
        data = fh.read()
        try:
            data = json.loads(data)
            plan = tfPlan.TerraformPlan(**data)
        except Exception as e:
            if opts.plan.endswith('.json'):
                print(f'Failed to load plan file from: {opts.template_path}')
                raise e
            plan = tfPlan.plan_from_stdout(data)

    if opts.subparser == 'check-no-new-access':
        checker=iamcheck_AccessAnalyzer.Comparator(account.Region, json.dumps(_load_json_file(opts.reference_policy)), opts.reference_policy_type)
        checker.run(plan)
        findings = checker.findings
    elif opts.subparser == 'check-access-not-granted':
        checker=iamcheck_AccessAnalyzer.AccessChecker(account.Region, opts.actions)
        checker.run(plan)
        findings = checker.findings
    else:
        validator=iamcheck_AccessAnalyzer.Validator(account.Account, account.Region, account.Partition)
        validator.run(plan)
        findings = validator.findings
        
    treat_as_blocking = opts.treat_as_blocking if 'treat_as_blocking' in vars(opts) else default_finding_types_that_are_blocking
    allowed_external_principals = opts.allowed_external_principals if 'allowed_external_principals' in vars(opts) else None
    reporter = Reporter(opts.ignore_finding, treat_as_blocking, allowed_external_principals)
    report = reporter.build_report_from(findings)
    LOGGER.info('Printing findings to the console...')
    report.print()
            
    if report.has_blocking_findings():
        exit(2)
    else:
        exit(0)


def cli_parse_opts():
    parent_parser = argparse.ArgumentParser(add_help=False)

    parent_parser.add_argument('--template-path', metavar="TEMPLATE_PATH", dest="template_path", required=True,
                        help='Terraform plan file (JSON)')
    parent_parser.add_argument('--region', dest="region", required=True, type=validate_region,
                        help="The region the resources will be deployed to.")
    parent_parser.add_argument('--profile', help='The named profile to use for AWS API calls.')
    parent_parser.add_argument("--config", nargs="+", help='Config file for running this script', action='append')
    parent_parser.add_argument('--enable-logging', help='Enable detailed logging.', default=False, action='store_true')
    parent_parser.add_argument('--exclude-resource-types', dest="exclude_resource_type", default=[],
                                 help='Resource types to exclude from parsing. Specify as a comma separated list of Terraform resource types. '
                                     'Please see README for full list of possible types.',
                                 action=ParseListFromCLI)

    parser = argparse.ArgumentParser(description='Parses IAM identity-based and resource-based policies from Terraform templates.')

    subparsers = parser.add_subparsers(dest='subparser')
    subparsers.required = True
    validate_parser = subparsers.add_parser('validate', help='Parses IAM identity-based and resource-based policies from Terraform templates '
                                                           'and runs them through IAM Access Analyzer for validation.  Returns the findings from '
                                                           'validation in JSON format.', parents=[parent_parser])
    validate_parser.add_argument('--ignore-finding', dest="ignore_finding", metavar='FINDING_CODE,RESOURCE_NAME,RESOURCE_NAME.FINDING_CODE',
                                 help='Allow validation failures to be ignored.\n'
                             'Specify as a comma separated list of findings to be ignored. Can be individual '
                             'finding codes (e.g. "PASS_ROLE_WITH_STAR_IN_RESOURCE"), a specific resource name '
                             '(e.g. "MyResource"), or a combination of both separated by a period.'
                             '(e.g. "MyResource.PASS_ROLE_WITH_STAR_IN_RESOURCE").',
                                 action=ParseFindingsToIgnoreFromCLI)
    validate_parser.add_argument('--treat-finding-type-as-blocking', dest="treat_as_blocking", metavar="ERROR,SECURITY_WARNING",
                                 help='Specify which finding types should be treated as blocking. Other finding types are treated '
                                     'as non-blocking. Defaults to "ERROR" and "SECURITY_WARNING". Specify as a comma separated '
                                     'list of finding types that should be blocking.  Possible values are "ERROR", '
                                     '"SECURITY_WARNING", "SUGGESTION", and "WARNING".  Pass "NONE" to ignore all errors.',
                                 default=default_finding_types_that_are_blocking, type=validate_finding_types_from_cli)

    validate_parser.add_argument('--allow-external-principals', dest='allowed_external_principals', metavar="ACCOUNT,ARN",
                                 help='A comma separated list of external principals that should be ignored.  Specify as '
                                     'a comma separated list of a 12 digit AWS account ID, a federated web identity '
                                     'user, a federated SAML user, or an ARN. Specify "*" to allow anonymous access. '
                                     '(e.g. 123456789123,arn:aws:iam::111111111111:role/MyOtherRole,graph.facebook.com)',
                                 action=ParseAllowExternalPrincipalsFromCLI)

    def add_policy_analysis_subparsers():
        # check-no-new-access command
        compare_parser = subparsers.add_parser('check-no-new-access', help='Parses IAM identity-based and resource-based policies from AWS Terraform templates '
                                                            'and runs them through IAM Access Analyzer for comparison with a reference policy.  Returns the response '
                                                            'in JSON format.', parents=[parent_parser])

        compare_parser.add_argument('--reference-policy', dest="reference_policy", required=True,
                                    help='Reference policy to be compared to.\n')

        compare_parser.add_argument('--reference-policy-type', dest="reference_policy_type", required=True,
                                    type=str, help='The type of the reference policy (identity or resource)')

        
        compare_parser.add_argument('--ignore-finding', dest="ignore_finding", metavar='FINDING_CODE,RESOURCE_NAME,RESOURCE_NAME.FINDING_CODE',
                                    help='Allow findings to be ignored.\n'
                                        'Specify as a comma separated list of findings to be ignored. Can be individual '
                                        'finding codes (e.g. "PASS_ROLE_WITH_STAR_IN_RESOURCE"), a specific resource name '
                                        '(e.g. "MyResource"), or a combination of both separated by a period.'
                                        '(e.g. "MyResource.PASS_ROLE_WITH_STAR_IN_RESOURCE").',
                                        action=ParseFindingsToIgnoreFromCLI)
        compare_parser.add_argument('--treat-findings-as-non-blocking', dest="findings_are_blocking", 
                                    help='If set, all findings will be treated as non-blocking',
                                    default=True, action='store_false')
        
        # check-access-not-granted command
        check_access_parser = subparsers.add_parser('check-access-not-granted', help='Parses IAM identity-based and resource-based policies from AWS Terraform templates '
                                                            'and runs them through IAM Access Analyzer to check that access to a list of actions is not granted.  Returns the response '
                                                            'in JSON format.', parents=[parent_parser])

        check_access_parser.add_argument('--actions', dest="actions", required=True,
                                    help= 'Actions that policies should not grant.'
                                    'Specify as a comma separated list of actions to be checked.'
                                    'The tool will make multiple requests if you provide more actions than the allowed quota.', action=ParseListFromCLI)
        
        check_access_parser.add_argument('--ignore-finding', dest="ignore_finding", metavar='FINDING_CODE,RESOURCE_NAME,RESOURCE_NAME.FINDING_CODE',
                                    help='Allow findings to be ignored.\n'
                                        'Specify as a comma separated list of findings to be ignored. Can be individual '
                                        'finding codes (e.g. "PASS_ROLE_WITH_STAR_IN_RESOURCE"), a specific resource name '
                                        '(e.g. "MyResource"), or a combination of both separated by a period.'
                                        '(e.g. "MyResource.PASS_ROLE_WITH_STAR_IN_RESOURCE").',
                                        action=ParseFindingsToIgnoreFromCLI)
        check_access_parser.add_argument('--treat-findings-as-non-blocking', dest="findings_are_blocking", 
                                    help='If set, all findings will be treated as non-blocking',
                                    default=True, action='store_false')
    
    add_policy_analysis_subparsers()

    args = parser.parse_args()

    #load yaml config
    if args.config is not None:
        for conf in [fileName for arg in args.config for fileName in arg]:
            LOGGER.debug(f'Config file: {conf}')
            config.load_config_yaml(conf, args.exclude_resource_type)

    #Make sure there is at least one policy to look for
    if len(config.iamPolicyAttributes) == 0:
        raise ValueError(f'No IAM policies defined!')
    
    set_profile(args.profile)
    validate_credentials(args.region)
    config.configure_logging(args.enable_logging)
    return args


if  __name__ == "__main__":
    try:
        main()
    except Exception as e:
        traceback.print_exc()
        print(f'ERROR: Unexpected error occurred. {str(e)}', file=sys.stderr)
        exit(1)


"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""
import datetime
import json
import logging
from iam_check.application_error import ApplicationError

LOGGER = logging.getLogger('iam-check')

def default_to_json(value):
	if isinstance(value, datetime.date):
		return value.isoformat()
	else:
		return value.__dict__

def _load_json_file(file_path):
    try:
        with open(file_path, 'r') as stream:
            data = stream.read()
            try:
                ret = json.loads(data)
            except Exception:
                logging.exception('Unable to parse json file. Invalid JSON detected.')
                raise ApplicationError('Unable to parse json file. Invalid JSON detected.')
    except FileNotFoundError:
        raise ApplicationError(f'File not found: {file_path}')

    return ret


class InvalidPolicyException(Exception):
	def __init__(self, message, policy):
		self.message = message
		self.policy = policy

	def to_string(self):
		return f'{self.message}\n{json.dumps(self.policy, default=default_to_json, indent=4)}'


import abc
import logging

logger = logging.getLogger(__name__)

class IamCheck(abc.ABC):
    def __init__(self, *args, **kwargs):
        name = type(self).__name__
        logger.debug(f'Initializing check: {name}')

        self.exceptions = kwargs.get('exceptions', [])
    
    @abc.abstractclassmethod
    def run(self, policy, **kwargs): pass



import logging
import yaml
import sys

# logging configuration
LOGGER = logging.getLogger('iam-policy-validator-for-terraform')

# AWS Account ID to use when unknown
awsAccount = '123456789012'

#IAM Policy checks to run
# The default is to run all checks if thhe list is empty
# iamChecks = []

#IAM policy resources
iamPolicyAttributes = {}

#Generate fake ARN
# default substitube is {<key>?<default>}
arnServiceMap = {}

validatePolicyResourceType = {}

def configure_logging(enable_logging):
    console_handler = logging.StreamHandler(sys.stdout)
    #console_handler.setLevel(logging.DEBUG)

    LOGGER.setLevel(logging.INFO)

    # log_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    console_handler.setFormatter(log_formatter)
    LOGGER.propagate = False
    # for handler in LOGGER.handlers:
    #     LOGGER.removeHandler(handler)
    LOGGER.addHandler(console_handler)
    if not enable_logging:
        LOGGER.disabled = True
def load_config_yaml(file, exclude_resource_type = []):
    global arnServiceMap
    global iamPolicyAttributes
    global validatePolicyResourceType
    
    with open(file, 'r') as fh:
        data = yaml.safe_load(fh)

    arnServiceMap = data.get('arnServiceMap', arnServiceMap)
    if 'arnServiceMap' in data:
        arnServiceMap = data['arnServiceMap'] 
        
    if 'iamPolicyAttributes' in data:
        iamPolicyAttributes = data['iamPolicyAttributes']
        for exclude_type in exclude_resource_type:
            del iamPolicyAttributes[exclude_type]
        
    if 'validatePolicyResourceType' in data:
        validatePolicyResourceType = data['validatePolicyResourceType']
    



import json
import logging
from typing import Optional, Union

LOGGER = logging.getLogger('iam-policy-validator-for-terraform')

class Policy:
    def __init__(self, **kwargs):
        self._version = ""
        self._id = ""
        self._statement = []

        accepted =['Version', 'Id', 'Statement', 'file', 'json']

        for key, value in kwargs.items():
            if key == 'Version':
                self.setVersion(value)
            elif key == 'Id':
                self.setId(value)
            elif key =='Statement':
                if isinstance(value, Statement):
                    self.addStatement(value)
                elif isinstance(value, dict):
                    self.addStatement(Statement(**value))
                elif isinstance(value, list):
                    for statement in value:
                        if isinstance(value, Statement):
                            self.addStatement(statement)
                        else:
                            self.addStatement(Statement(**statement))
                else:
                    raise ValueError(f'Statements expects list or dictionary')

            elif key == 'file':
                if len(kwargs) != 1:
                    raise ValueError('Invalid parameter: file can not be used with other arguments')
                with open(value, 'r') as fh:
                    data = json.load(fh)
                self.__init__(**data)

            elif key == 'json':
                if len(kwargs) != 1:
                    raise ValueError('Invalid parameter: file can not be used with other arguments')
                data = json.loads(value)
                self.__init__(**data)

            else:
                raise ValueError(f'Invalid parameter: {key} not a valid parameter')

    def __eq__(self, o) -> bool:
        """Overload = operator to test if two policies are the same"""
        if self._version != o._version:
            return False
        if self._id != o._id:
            return False
        if len(self._statement) != len(o._statement):
            return False
        for s in self._statement:
            if s not in o._statement:
                return False
        return True

    def __str__(self) -> str:
        """return a JSON representative of the iam policy"""
        obj = {}
        if self._version != "":
            obj['Version'] = self._version
        
        if self._id != "":
            obj['Id'] = self._id

        if len(self._statement) == 1:
            t = json.loads(str(self._statement[0]))
            obj['Statement'] = t
        else:
            obj['Statement'] = []
            for statement in self._statement:
                t = json.loads(str(statement))
                obj['Statement'].append(t)
        return json.dumps(obj, indent=4)

    def setId(self, id: str) -> None:
        self._id = id

    def setVersion(self, version: str) -> None:
        valid = ["2012-10-17", "2008-10-17"]

        if version not in valid:
            raise ValueError(f'Invalid IAM policy version: {version}')

        self._version = version

    def addStatement(self, statement) -> None:
        if statement not in self._statement:
            self._statement.append(statement)

    def deleteStatement(self, statement) -> None:
        self._statement = [s for s in self._statement if s != statement]

    def getId(self) -> str:
        return self._id

    def getVersion(self) -> str:
        return self._version

    def getStatement(self) -> str:
        return self._statement

class Statement:
    def __init__(self, json: str= None, **kwargs):
        self._sid = None
        self._principal = None
        self._effect = ""
        self._action = []
        self._resource = []
        self._condition = None

        self._notAction = False
        self._notPrincipal = False
        self._notResource = False

        if json is not None:
            data = json.loads(json)
            self.__init__(**data)
            return

        if 'Action' in kwargs and 'NotAction' in kwargs:
            raise ValueError('Parameter: Can not use Action with NotAction')
        if 'Principal' in kwargs and 'NotPrincipal' in kwargs:
            raise ValueError('Parameter: Can not use Action with NotPrincipal')
        if 'Resource' in kwargs and 'NotResource' in kwargs:
            raise ValueError('Parameter: Can not use Action with NotResource')

        for arg, value in kwargs.items():
            if arg == 'Sid':
                self.setSid(value)
            elif arg == 'Effect':
                self.setEffect(value)
            elif arg == 'Action' or arg == 'NotAction':
                if isinstance(value, list):
                    for action in value:
                        self.addAction(action, arg == 'NotAction')
                else:
                    self.addAction(value, arg == 'NotAction')
            elif arg == 'Principal' or arg == 'NotPrincipal' :
                self.setPrincipal(value, arg == 'NotPrincipal')
            elif arg == 'Resource' or arg == 'NotResource':
                if isinstance(value, list):
                    for resource in value:
                        self.addResource(resource, arg == 'NotResource')
                else:
                    self.addResource(value, arg == 'NotResource')
            elif arg == 'Condition':
                self._condition = value
            else:
                raise ValueError(f'Invalid parameter: {arg} not a valid parameter')
            
    def __eq__(self, o) -> bool:
        if self._notAction != o._notAction:
            return False
        if self._notPrincipal != o._notPrincipal:
            return False
        if self._notResource != o._notResource:
            return False
        if self._sid != o._sid:
            return False
        if self._effect != o._effect:
            return False
        if len(self._action) != len(o._action):
            return False
        if len(self._resource) != len(o._resource):
            return False
        
        me = sorted(self._resource)
        you = sorted(o._resource)
        if me != you:
            return False
        
        me = sorted(self._action)
        you = sorted(o._action)
        if me != you:
            return False      

        me = json.dumps(self._principal, sort_keys = True)
        me = json.dumps(o._principal, sort_keys = True)
        if me != you:
            return False      

        me = json.dumps(self._condition, sort_keys = True)
        me = json.dumps(o._condition, sort_keys = True)
        if me != you:
            return False      

        return True

    def __str__(self) -> str:
        obj = {}
        if self._sid is not None:
            obj['Sid'] = self._sid
        if self._principal is not None:
            key = 'NotPrincipal' if self._notPrincipal else 'Principal'
            obj[key] = self._principal
        if self._effect != "":
            obj['Effect'] = self._effect
        if self._condition is not None:
            obj['Condition'] = self._condition
        
        key = 'NotAction' if self._notAction else 'Action'
        if len(self._action) == 1:
            obj[key] = self._action[0]
        else:
            obj[key] = self._action

        key = 'NotResource' if self._notResource else 'Resource'
        if len(self._resource) == 1:
            obj[key] = self._resource[0]
        else:
            obj[key] = self._resource
        
        return json.dumps(obj, indent=4, default=str)


    def setSid(self, sid: Optional[str] = None):
        self._sid = sid

    def setEffect(self, effect: Optional[str] = None):
        valid = [None, "Allow", "Deny"]
        if effect not in valid:
            raise ValueError(f'Invalid effect: {effect}')
        self._effect = effect
    
    def setPrincipal(self, principal: Union[str, dict, None]= None, NotPrincipal = False):
        if self._principal is not None and self.NotPrincipal != NotPrincipal:
            raise ValueError(f'Can not use Principal with NotPrincipal')
        self._notPrincipal = NotPrincipal
        self._principal = principal

    def addAction(self, action: str, notAction = False):
        if len(self._action) > 0 and self._notAction != notAction:
            raise ValueError(f'Can not use Action with NotAction')
        self._notAction = notAction
        self._action.append(action)

    def addResource(self, resource: str, notResource = False):
        if len(self._resource) > 0 and self._notResource != notResource:
            raise ValueError(f'Can not use Resource with NotResource')
        self._notResource = notResource
        if resource not in self._resource:
            self._resource.append(resource)

    def deleteAction(self, action: str):
        self._action = [a for a in self._action if a != action]

    def deleteResource(self, resource: str):
        self._resource = [r for r in self._resource if r != resource]
    
    def getSid(self):
        return self._sid
    
    def getEffect(self):
        return self._effect

    def getAction(self):
        if self._notAction:
            return None
        return self._action
    
    def getNotAction(self):
        if self._notAction:
            return self._action
        return None

    def getPrincipal(self):
        if self._notPrincipal:
            return None
        return self._principal

    def listPrincipal(self, notPrincipal = False):
        if notPrincipal != self._notPrincipal:
            return [None]
        principal = self._principal

        if principal is None:
            return [None]

        if principal == "*":
            return ["*"]

        principals = []
        for svc,ids in principal.items():
            if isinstance(ids,str):
                principals.append(f'{svc}:{ids}')
                continue
            for id in ids:
                principals.append(f'{svc}:{ids}')
        return principals

    def getNotPrincipal(self):
        if self._notPrincipal:
            return self._principal
        return None
    
    def getResource(self):
        if self._notResource:
            return None
        return self._resource 

    def getNotResource(self):
        if self._notResource:
            return self._resource 
        return None

    def getCondition(self):
        if self._condition:
            return self._condition
        return None


import config
import json
import pytest
import lib.tfPlan as tfPlan

class TestTerraformResource:
    def test_core_function(self):
        err = []
        data = {
            'address': 'foo.bar',
            'name': 'bar',
            'values': {
                'baz': 'Hello!'
            },
            'sensitive_values':{
                'tags_all': {}
            }
        }
        match = '{\n    "address": "foo.bar",\n    "name": "bar",\n    "values": {\n        "baz": "Hello!"\n    },\n    "sensitive_values": {\n        "tags_all": {}\n    }\n}'
        x = tfPlan.TerraformResource(**data)
        if str(x) != match:
            err.append('Invalid JSON representation')

        if x.getValue('baz') != 'Hello!':
            err.append('Can not get value of baz')
        

        if x.getValue('tags_all') != {}:
            err.append('Can not get value of tags_all')

        x.setValue('big', 'shot')
        if x.getValue('big') != 'shot':
            err.append('Failed setting a value with setqVaule()')
        assert( err == [])




from lib import _load_json_file
from lib.reporter import Reporter
from lib.tfPlan import TerraformPlan
from lib.iamcheck_AccessAnalyzer import Comparator, Validator
from iam_check.config import load_config_yaml
import pytest
import json

class TestAccessAnalyzer:

    load_config_yaml("config/default.yaml", exclude_resource_type = [])

    def test_a2_policy_validation(self):
        file = _load_json_file("test/iam_policy/test_plan.json")
        plan = TerraformPlan(**file)
        check = Validator("123456789012", "us-west-2", "aws")
        check.run(plan)
        findings = _load_json_file("test/iam_policy/findings.json")
        assert(Reporter(None, None, None).build_report_from(check.findings).to_json() == findings)

    def test_a2_policy_check_no_new_access_mixed_policies(self):
        for dir_name in ["multiple_policies", "role_inline_assume_policy"]:
            file = _load_json_file(f"test/{dir_name}/test_plan.json")
            plan = TerraformPlan(**file)
            for policy_type in ["resource", "identity"]:
                reference_policy = _load_json_file(f"test/{dir_name}/{policy_type}_reference_policy.json")
                check = Comparator("us-west-2", json.dumps(reference_policy), reference_policy_type=f'{policy_type}')
                check.run(plan)
                test_findings = _load_json_file(f"test/{dir_name}/{policy_type}_findings.json")
                assert(Reporter(None, ['ERROR', 'SECURITY_WARNING'], None).build_report_from(check.findings).to_json() == test_findings)
    
    def test_a2_policy_check_no_new_access_identity_policies(self):
        for dir_name in ["identity_policies"]:
            file = _load_json_file(f"test/{dir_name}/test_plan.json")
            plan = TerraformPlan(**file)
            for policy_type in ["identity"]:
                reference_policy = _load_json_file(f"test/{dir_name}/{policy_type}_reference_policy.json")
                check = Comparator("us-west-2", json.dumps(reference_policy), reference_policy_type=f'{policy_type}')
                check.run(plan)
                test_findings = _load_json_file(f"test/{dir_name}/{policy_type}_findings.json")
                assert(Reporter(None, ['ERROR', 'SECURITY_WARNING'], None).build_report_from(check.findings).to_json() == test_findings)